from .base import BasePlugin
from ..constants import PluginHook


class LegacyEnvPlugin(BasePlugin):
    """
    Plugin that adds legacy style (REDIS_1_PORT_6379_TCP_ADDR) environment variables to containers based on links.
    """

    def load(self):
        self.add_hook(PluginHook.PRE_RUN_CONTAINER, self.add_link_envs)

    def add_link_envs(self, host, instance, task):
        """
        Takes the instance and modifies the environment to have legacy link variables.
        """
        for alias, target in instance.links.items():
            # Ask Docker for all open ports
            ports = host.client.inspect_container(target.name)['NetworkSettings']['Ports']
            if ports:
                for port, _ in ports.items():
                    number, protocol = port.split("/")
                    name = "{}_1_PORT_{}_{}".format(
                        alias.replace("-", "_").upper(),
                        number,
                        protocol.upper(),
                    )
                    instance.environment[name] = "{}://{}:{}".format(protocol, alias, number)
                    instance.environment[name + "_ADDR"] = alias
                    instance.environment[name + "_PORT"] = number



import os
import logging
import subprocess

from .base import BasePlugin
from ..cli.tasks import Task
from ..constants import PluginHook
from ..exceptions import BuildFailureError


class BuildScriptsPlugin(BasePlugin):

    def load(self):
        self.add_hook(PluginHook.PRE_BUILD, self.run_pre_build_script)
        self.add_hook(PluginHook.POST_BUILD, self.run_post_build_script)
        self.add_hook(PluginHook.PRE_RUN_CONTAINER, self.run_pre_start_script)

    def run_pre_build_script(self, host, container, task):
        """
        Runs the pre build scripts.
        """
        self.run_script("pre-build", container, task)

    def run_post_build_script(self, host, container, task):
        """
        Runs the post build scriipts.
        """
        self.run_script("post-build", container, task)

    def run_pre_start_script(self, host, instance, task):
        """
        Runs the post build scriipts.
        """
        self.run_script("pre-start", instance.container, task)

    def run_script(self, name, container, task):
        """
        Runs a script, logs its output, and errors if it breaks.

        We call interpreters directly as the scripts may not always be +x and so
        we cannot call them directly and rely on their shebang line.
        """
        for script_extension, interpreter in [(".sh", "bash"), (".py", "python")]:
            script_path = os.path.join(container.path, name + script_extension)
            if os.path.exists(script_path):
                # Make a build directory, removing any old one if it exists
                build_dir = os.path.join(container.path, "build")
                if os.path.isdir(build_dir):
                    subprocess.call(["rm", "-rf", build_dir])
                os.mkdir(build_dir)
                # Run the script
                script_task = Task("Running {}".format(name), parent=task, collapse_if_finished=True)
                logger = logging.getLogger('build_logger')
                process = subprocess.Popen(
                    [interpreter, script_path],
                    cwd=container.path,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT
                )
                while True:
                    line = process.stdout.readline().rstrip().decode("utf8")
                    logger.info(line)
                    if line.strip():
                        script_task.set_extra_info(
                            script_task.extra_info[-3:] + [line]
                        )
                    if not line and process.poll() is not None:
                        break
                exit_code = process.wait()
                if exit_code:
                    script_task.finish(status="Failed", status_flavor=Task.FLAVOR_BAD)
                    raise BuildFailureError("Script {} failed".format(name))
                else:
                    script_task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)
                break



import sys
import click

from .base import BasePlugin
from ..cli.colors import RED
from ..cli.tasks import Task
from ..constants import PluginHook
from ..docker.introspect import FormationIntrospector
from ..docker.runner import FormationRunner
from ..exceptions import BadConfigError, ImageNotFoundException


class BootPlugin(BasePlugin):
    """
    Plugin for running certain containers at boot.

    Containers should have a "boot:" config key with "build" and "run" sub keys
    listing the containers to boot before the build or run phase, respectively,
    and if they need them (required) or can use them if available (optional):

    boot:
        build:
            apt-cacher: optional
            ssh-agent: required
        run:
            ssh-agent: required

    These values are inherited from build parents.
    """

    provides = ["boot-containers"]

    def load(self):
        self.add_hook(PluginHook.PRE_BUILD, self.pre_build)
        self.add_hook(PluginHook.PRE_RUN_CONTAINER, self.pre_start)

    def pre_build(self, host, container, task):
        boot_containers = self.calculate_boot_containers("build", container)
        self.run_boot_containers(host, boot_containers, task)

    def pre_start(self, host, instance, task):
        boot_containers = self.calculate_boot_containers("run", instance.container)
        self.run_boot_containers(host, boot_containers, task)

    def calculate_boot_containers(self, phase, container):
        """
        Given a Container object and phase being "build" or "run",
        calculates all boot containers needed based on build parents.
        Returns a dict of {container: required}
        """
        configs = container.get_ancestral_extra_data("boot")
        boot_container_names = {}
        for config in configs:
            for name, required in config.get(phase, {}).items():
                required = required.lower().strip() == "required"
                # Any required value overrides a previous optional value
                boot_container_names[name] = required or boot_container_names.get(name, False)
        # Translate the names into container objects
        try:
            boot_containers = {
                self.app.containers[name]: required
                for name, required in boot_container_names.items()
            }
        except KeyError as e:
            raise BadConfigError("Invalid boot container specification: {}".format(e))
        # Check none of them are circular dependencies
        ancestry = self.app.containers.build_ancestry(container) + [container]
        for boot_container, required in boot_containers.items():
            if boot_container in ancestry and required:
                raise BadConfigError(
                    "Container {} has a boot container of {}, which is in its own build ancestry".format(
                        container.name,
                        boot_container.name,
                    )
                )
        return boot_containers

    def run_boot_containers(self, host, containers, task):
        """
        Takes a dict of {container: required} and makes sure they're all running
        (required ones must be or an error is raised; optional ones are if they
        are available locally).
        """
        formation = FormationIntrospector(host, self.app.containers).introspect()
        to_boot = set()
        for container, required in containers.items():
            # See if container is already running
            if any(instance.container == container for instance in formation):
                continue
            # See if it can be started (use a runner just for missing images - maybe improve this)
            try:
                formation.add_container(container, host)
            except ImageNotFoundException:
                if required:
                    click.echo(RED("Cannot launch required boot container {} - no image".format(container.name)))
                    sys.exit(1)
            else:
                to_boot.add(container)
        # Boot those containers
        if to_boot:
            boot_task = Task("Running boot containers", parent=task)
            formation = FormationIntrospector(host, self.app.containers).introspect()
            for container in to_boot:
                formation.add_container(container, host)
            runner = FormationRunner(self.app, host, formation, boot_task, stop=False)
            runner.run()
            boot_task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)



import dockerpty
import functools
import os
import sys
import threading
import time

from docker.errors import NotFound

from .introspect import FormationIntrospector
from .towline import Towline
from ..cli.tasks import Task
from ..constants import PluginHook
from ..exceptions import ContainerBootFailure, DockerRuntimeError, DockerInteractiveException, NotFoundException
from ..utils.sorting import dependency_sort
from ..utils.threading import ExceptionalThread, ThreadSet


network_lock = threading.Lock()

# Tracks which containers are being started/stopped globally to avoid starting the same one twice.
changing_containers = ThreadSet()


class FormationRunner:
    """
    Takes a ContainerFormation to aim for and a host to run it on, and brings
    the two in line by starting/stopping/configuring containers.

    It can run actions in parallel in background threads if needs be.
    """

    def __init__(self, app, host, formation, task, stop=True):
        self.app = app
        self.host = host
        self.formation = formation
        self.introspector = FormationIntrospector(self.host, self.formation.graph)
        self.task = task
        # Allows things to override and not have anything stop
        self.stop = stop

    def run(self):
        """
        Runs through and performs all the actions. Blocks until completion.
        """
        self.actions = []
        # Check the formation is valid
        self.formation.validate()
        # Work out what containers need turning off, and which need turning on
        # Containers that have changes will need both.
        to_stop = set()
        to_start = set()
        current_formation = self.introspector.introspect()
        for instance in current_formation:
            if instance not in self.formation:
                to_stop.add(instance)
        # Now see if there are any that are entirely new
        for instance in self.formation:
            if instance not in current_formation:
                to_start.add(instance)
            else:
                # It's in both - stop and start if it's changed
                if instance.different_from(current_formation[instance.name]):
                    to_stop.add(instance)
                    to_start.add(instance)
        # Stop containers in parallel
        if to_stop and self.stop:
            self.stop_containers(to_stop)
        # Start containers in parallel
        if to_start:
            self.start_containers(to_start)

    # Shared "dependency-based parallel execution" code

    def parallel_execute(self, instances, ready_to_execute, executor, done=None):
        """
        Runs the "executor" in parallel threads on "instances" when the condition
        "ready_to_execute" is met for an instance. Handles deadlocking as well.
        """
        idle_iterations = 0
        queued = set(instances)
        processing = set()
        done = done or set()
        threads = {}
        while queued or processing:
            # See if we can stop anything new - everything that depends on it must also be stopped
            for instance in list(queued):
                if ready_to_execute(instance, done):
                    threads[instance] = ExceptionalThread(
                        target=executor,
                        args=(instance,),
                        daemon=True,
                    )
                    threads[instance].start()
                    queued.remove(instance)
                    processing.add(instance)
                    idle_iterations = 0
            # See if anything finished stopping
            for instance in list(processing):
                if not threads[instance].is_alive():
                    processing.remove(instance)
                    done.add(instance)
                    # Collect exceptions from the thread - if it's an interactive exception, run the rest of it.
                    try:
                        threads[instance].maybe_raise()
                    except DockerInteractiveException as e:
                        e.handler()
                        sys.exit(0)
                    del threads[instance]
                    idle_iterations = 0
            # If there's nothing in progress, we've deadlocked
            if idle_iterations > 10 and queued and not processing:
                raise DockerRuntimeError(
                    "Deadlock during stop: Cannot stop any of {}.".format(
                        ", ".join(i.name for i in queued),
                    ),
                )
            idle_iterations += 1
            # Don't idle hot
            time.sleep(0.1)

    # Stopping

    def stop_containers(self, instances):
        """
        Stops all the specified containers in parallel, still respecting links
        """
        current_formation = self.introspector.introspect()

        # Inner function that we can pass to dependency_sort
        @functools.lru_cache(maxsize=512)
        def get_incoming_links(instance):
            result = set()
            for potential_linker in current_formation:
                links_to = potential_linker.links.values()
                if instance in links_to:
                    result.add(potential_linker)
            return result

        # Resolve container list to include descendency
        instances = dependency_sort(instances, get_incoming_links)
        # Parallel-stop things
        self.parallel_execute(
            instances,
            lambda instance, done: all((linker in done) for linker in get_incoming_links(instance)),
            executor=self.stop_container,
        )

    def stop_container(self, instance):
        # Wait for the global container manipulation lock
        with changing_containers.entry_lock(instance.name):
            # See if it was already stopped
            if not self.host.container_running(instance.name, ignore_exists=True):
                return
            # Stop the container
            stop_task = Task(
                "Stopping {}".format(instance.container.name),
                parent=self.task,
                collapse_if_finished=True,
            )
            self.host.client.stop(
                instance.name,
                timeout=0 if instance.container.fast_kill else 10,
            )
            stop_task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

    # Starting

    def start_containers(self, instances):
        """
        Starts all the specified containers in parallel, respecting links
        """
        current_formation = self.introspector.introspect()
        self.parallel_execute(
            instances,
            lambda instance, done: all((dependency in done) for dependency in instance.links.values()),
            executor=self.start_container,
            done=set(started_instance for started_instance in current_formation),
        )

    def remove_stopped(self, instance):
        """
        Sees if there is a container with the same name and removes it if
        there is and it's stopped.
        """
        if self.host.container_exists(instance.name):
            if self.host.container_running(instance.name):
                raise DockerRuntimeError("The container {} is already running.".format(instance.container.name))
            else:
                self.host.client.remove_container(instance.name)

    def start_container(self, instance):
        """
        Creates the Docker container on the host, ready to be started.
        """
        # Make sure it's not an abstract container being started.
        if instance.container.abstract and not instance.foreground:
            raise ValueError("You cannot boot an abstract container.")

        # Wait for the global container manipulation lock
        with changing_containers.entry_lock(instance.name):
            # See if the container was already started
            if self.host.container_running(instance.name, ignore_exists=True):
                return

            start_task = Task(
                "Starting {}".format(instance.container.name),
                parent=self.task,
                collapse_if_finished=True,
            )

            self.remove_stopped(instance)

            # Run plugins
            self.app.run_hooks(PluginHook.PRE_START, host=self.host, instance=instance, task=start_task)

            # See if network exists and if not, create it
            with network_lock:
                try:
                    self.host.client.inspect_network(instance.formation.network)
                except NotFound:
                    self.host.client.create_network(
                        name=instance.formation.network,
                        driver="bridge",
                    )

            # Create network configuration for the new container
            networking_config = self.host.client.create_networking_config({
                instance.formation.network: self.host.client.create_endpoint_config(
                    aliases=[instance.formation.network],
                    links=[
                        (link.name, alias)
                        for alias, link in instance.links.items()
                    ]
                ),
            })

            # Work out volumes configuration
            volume_mountpoints = []
            volume_binds = {}

            def add_volume_mount(mount_path, volume):
                if self.host.supports_cached_volumes and ",cached" not in volume.mode:
                    volume.mode = volume.mode + ",cached"
                volume_mountpoints.append(mount_path)
                volume_binds[volume.source] = {"bind": mount_path, "mode": volume.mode}

            for mount_path, volume in instance.container.bound_volumes.items():
                if os.path.isdir(volume.source) or os.environ.get("BAY_VOLUME_HOME"):
                    add_volume_mount(mount_path, volume)
                elif volume.required:
                    raise NotFoundException(
                        "Volume mount source directory {} does not exist".format(volume.source)
                    )
            # Add any active devmodes
            for mount_name in instance.devmodes:
                for mount_path, volume in instance.container.devmodes[mount_name].items():
                    if os.path.isdir(volume.source) or os.environ.get("BAY_VOLUME_HOME"):
                        add_volume_mount(mount_path, volume)
                    else:
                        raise NotFoundException(
                            "Devmode source directory {} does not exist".format(volume.source)
                        )
            for mount_path, volume in instance.container.named_volumes.items():
                add_volume_mount(mount_path, volume)

            # Create container
            container_pointer = self.host.client.create_container(
                instance.image_id,
                command=instance.command,
                detach=not instance.foreground,
                stdin_open=instance.foreground,
                tty=instance.foreground,
                # Ports is a list of ports in the container to expose
                ports=list(instance.ports.keys()),
                environment=instance.environment,
                volumes=volume_mountpoints,
                name=instance.name,
                host_config=self.host.client.create_host_config(
                    mem_limit=instance.mem_limit,
                    binds=volume_binds,
                    port_bindings=instance.ports,
                    publish_all_ports=True,
                    security_opt=['seccomp:unconfined'],
                ),
                networking_config=networking_config,
                labels={
                    "com.eventbrite.bay.container": instance.container.name,
                }
            )
            try:
                # Foreground containers launch into a PTY at this point. We use an exception so that
                # it happens in the main thread.
                if instance.foreground:
                    def handler():
                        dockerpty.start(self.host.client, container_pointer)
                        self.host.client.remove_container(container_pointer)
                    start_task.finish(status="Going to shell", status_flavor=Task.FLAVOR_GOOD)
                    raise DockerInteractiveException(handler)

                else:
                    # Make a towline instance and wait on it
                    self.host.client.start(container_pointer)
                    towline = Towline(self.host, instance.name)
                    while True:
                        status, message = towline.status
                        if status is None:
                            if message is not None:
                                start_task.update(status=message)
                        elif status is True:
                            break
                        elif status is False:
                            raise ContainerBootFailure(
                                "Failed during towline",
                                instance=instance,
                            )
                        time.sleep(0.5)

                try:
                    # Replace the instance with an introspected copy of the live one so it has networking details
                    instance = FormationIntrospector(
                        self.host,
                        self.app.containers,
                    ).introspect_single_container(instance.name)
                except DockerRuntimeError:
                    raise ContainerBootFailure(
                        "Failed after towline",
                        instance=instance,
                    )

                # Run plugins
                self.app.run_hooks(PluginHook.POST_CONTAINER_START, host=self.host, instance=instance, task=start_task)
                self.app.run_hooks(PluginHook.POST_SERVICE_START, host=self.host, instance=instance, task=start_task)

            except ContainerBootFailure as e:
                message = "{}\n\n{}".format(
                    "Container {} failed to boot! ({})".format(e.instance.container.name, e.message),
                    self.host.client.logs(e.instance.name, tail=10).decode('utf-8'),
                )
                raise DockerRuntimeError(
                    message,
                    code="BOOT_FAIL",
                    instance=e.instance,
                )

            start_task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)



from collections import defaultdict

import attr
import click

from .base import BasePlugin
from ..cli.argument_types import ContainerType, HostType, MountType
from ..cli.colors import CYAN, GREEN, PURPLE
from ..cli.tasks import Task


@attr.s
class DevModesPlugin(BasePlugin):
    """
    Plugin for managing dev checkouts in a container.
    """

    requires = ["up"]

    def load(self):
        self.add_command(mounts)
        self.add_command(mount)
        self.add_command(unmount)
        self.add_alias(unmount, "umount")


@click.command()
@click.option("--profile-only", "-p", default=False, is_flag=True)
@click.pass_obj
def mounts(app, profile_only):
    """
    List all current dev mounts.
    """
    dev_mounts = defaultdict(dict)

    for container in app.containers:
        unmounted_devmodes = set(container.devmodes.keys())
        runtime_options = app.containers.options(container)
        if runtime_options:
            devmodes = app.containers.options(container).get('devmodes')
            dev_mounts[container.name]['mounted'] = sorted(devmodes)
            dev_mounts[container.name]['unmounted'] = unmounted_devmodes.difference(devmodes)
        elif not profile_only:
            # only containers in profile have runtime_options set, so if profile_only
            # then skip ones with no options altogether regrdless is the have devmodes
            dev_mounts[container.name]['mounted'] = []
            dev_mounts[container.name]['unmounted'] = unmounted_devmodes

    for name, devmodes in dev_mounts.items():
        if devmodes['mounted']:
            click.echo('{}: \nMounted: {}\nUnmounted: {}'.format(
                CYAN(name),
                GREEN(', '.join(sorted(devmodes['mounted']))),
                PURPLE(', '.join(sorted(devmodes['unmounted']))),
            ))
        elif devmodes['unmounted']:
            click.echo('{}: \nUnmounted: {}'.format(
                CYAN(name),
                PURPLE(', '.join(sorted(devmodes['unmounted']))),
            ))


@click.command()
@click.argument('mount', type=MountType())
@click.argument('container', default='all', type=ContainerType(all=True))
@click.option("--host", "-h", type=HostType(), default="default")
@click.option('--up/--no-up', '-u', default=True)
@click.pass_obj
def mount(app, mount, container, host, up):
    """
    Mount a dev checkout in a given container.
    """
    # Check the profile is loaded
    if app.user_profile.file_path is None:
        click.echo("No profile loaded. Please select a profile using `bay profile <profile_name>`")
        return
    # Add the devmode
    mutate_devmounts(app, container, add=[mount])

    if up:
        # Run tug up to apply any changes
        app.invoke("up")


@click.command()
@click.argument('mount', type=MountType())
@click.argument('container', default='all', type=ContainerType(all=True))
@click.option("--host", "-h", type=HostType(), default="default")
@click.option('--up/--no-up', '-u', default=True)
@click.pass_obj
def unmount(app, mount, container, host, up):
    """
    Unmount a dev checkoutin a given container.
    """
    # Check the profile is loaded
    if app.user_profile.file_path is None:
        click.echo("No profile loaded. Please select a profile using `bay profile <profile_name>`")
        return
    # Remove the devmode
    mutate_devmounts(app, container, remove=[mount])

    if up:
        # Run tug up to apply any changes
        app.invoke("up")


def mutate_devmounts(app, containers, add=None, remove=None):
    """
    Applies devmount changes to a set of containers in a profile.
    """
    # Give add/remove good defaults
    add = add or []
    remove = remove or []

    # Convert any single instance of a container into a singleton list
    if not isinstance(containers, list):
        containers = [containers]
    else:
        containers = containers

    # For each provided container, try mounting
    profile = app.user_profile
    changed_containers = set()
    # Mounts to add
    for mount in add:
        task = Task("Mounting {}".format(mount), parent=app.root_task)
        for con in sorted(containers, key=lambda c: c.name):
            if mount in con.devmodes and mount not in profile.containers.get(con.name, {}).get('devmodes', set()):
                subtask = Task(con.name, parent=task)
                if not profile.containers.get(con.name):
                    profile.containers[con.name] = {}
                if not profile.containers[con.name].get('devmodes'):
                    profile.containers[con.name]['devmodes'] = set()
                profile.containers[con.name]['devmodes'].add(mount)
                changed_containers.add(con)
                subtask.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)
    # Mounts to remove
    for mount in remove:
        task = Task("Unmounting {}".format(mount), parent=app.root_task)
        for con in sorted(containers, key=lambda c: c.name):
            if mount in con.devmodes and mount in profile.containers.get(con.name, {}).get('devmodes', set()):
                subtask = Task(con.name, parent=task)
                profile.containers[con.name]['devmodes'].remove(mount)
                changed_containers.add(con)
                subtask.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

    if changed_containers:
        profile.save()
        profile.apply(app.containers)



import click
import sys

from .base import BasePlugin
from ..cli.colors import RED


class HelpPlugin(BasePlugin):
    """
    Just makes "bay help" show the help
    """

    def load(self):
        self.add_command(help)


@click.command()
@click.argument("command_name", default=None, required=False)
@click.pass_context
def help(ctx, command_name):
    """
    Shows main command list.
    """
    from ..cli import cli
    # Find subcommand
    if command_name:
        subcommand = cli.get_command(None, command_name)
        if subcommand is None:
            click.echo(RED("There is no command {}".format(command_name)))
            sys.exit(1)
        else:
            # Override info name so help prints correctly
            ctx.info_name = subcommand.name
            click.echo(subcommand.get_help(ctx))
    # Print main help
    else:
        ctx.info_name = "bay"
        ctx.parent = None
        click.echo(cli.get_help(ctx))



import attr
import click
import sys

from .base import BasePlugin
from ..cli.argument_types import ContainerType, HostType
from ..cli.colors import CYAN, RED
from ..cli.tasks import Task
from ..constants import PluginHook
from ..docker.introspect import FormationIntrospector
from ..docker.runner import FormationRunner
from ..exceptions import DockerRuntimeError, ImageNotFoundException


@attr.s
class RunPlugin(BasePlugin):
    """
    Plugin for running containers.
    """

    requires = ["tail"]

    def load(self):
        self.add_command(run)
        self.add_alias(run, "start")
        self.add_command(shell)
        self.add_command(stop)
        self.add_command(restart)
        self.add_alias(restart, "hup")
        self.add_alias(restart, "reload")


@click.command()
@click.argument("containers", type=ContainerType(), nargs=-1)
@click.option("--host", "-h", type=HostType(), default="default")
@click.option("--tail/--notail", "-t", default=False)
@click.pass_obj
def run(app, containers, host, tail):
    """
    Runs containers by name, including any dependencies needed
    """
    # Get the current formation
    formation = FormationIntrospector(host, app.containers).introspect()
    # Make a Formation that represents what we want to do by taking the existing
    # state and adding in the containers we want
    for container in containers:
        try:
            formation.add_container(container, host)
        except ImageNotFoundException as e:
            # If it's the container we're trying to add directly, have one error -
            # otherwise, say it's a link
            if e.image == container.image_name:
                click.echo(RED(
                    "This container ({name}) does not have a built image. Try `bay build {name}` first.".format(
                        name=container.name,
                    )
                ))
                sys.exit(1)
            elif hasattr(e, "container"):
                click.echo(RED("No image for linked container {name} - try `bay build {name}` first.".format(
                    name=e.container.name,
                )))
                sys.exit(1)
            else:
                click.echo(RED("No image for linked container %s!" % e.image))
                sys.exit(1)
    # Run that change
    task = Task("Starting containers", parent=app.root_task)
    run_formation(app, host, formation, task)
    # If they asked to tail, then run tail
    if tail:
        if len(containers) != 1:
            click.echo(RED("You cannot tail more than one container!"))
            sys.exit(1)
        app.invoke("tail", host=host, container=containers[0], follow=True)


@click.command()
@click.argument("container", type=ContainerType())
@click.option("--host", "-h", type=HostType(), default="default")
@click.argument("command", nargs=-1, default=None)
@click.pass_obj
def shell(app, container, host, command):
    """
    Runs a single container with foreground enabled and overridden to use bash.
    """
    # Get the current formation
    formation = FormationIntrospector(host, app.containers).introspect()
    # Make a Formation with that container launched with bash in foreground
    try:
        instance = formation.add_container(container, host)
    except ImageNotFoundException as e:
        click.echo(RED(str(e)))
        sys.exit(1)
    instance.foreground = True
    if command:
        instance.command = ['/bin/bash -lc "{}"'.format(' '.join(command))]
    else:
        instance.command = ["/bin/bash -l"]
    # Run that change
    task = Task("Shelling into {}".format(container.name), parent=app.root_task)
    run_formation(app, host, formation, task)


@click.command()
@click.argument("containers", type=ContainerType(), nargs=-1)
@click.option("--host", "-h", type=HostType(), default="default")
@click.pass_obj
def stop(app, containers, host):
    """
    Stops containers and ones that depend on them
    """
    formation = FormationIntrospector(host, app.containers).introspect()
    # Look through the formation and remove the containers matching the name
    for instance in list(formation):
        # If there are no names, then we remove everything
        if instance.container in containers or (not containers and not instance.container.system):
            # Make sure that it was not removed already as a dependent
            if instance.formation:
                formation.remove_instance(instance)
    # Run the change
    task = Task("Stopping containers", parent=app.root_task)
    run_formation(app, host, formation, task)


@click.command()
@click.argument("containers", type=ContainerType(), nargs=-1)
@click.option("--host", "-h", type=HostType(), default="default")
@click.pass_obj
def restart(app, containers, host):
    """
    Stops and then starts containers.
    """
    app.invoke("stop", containers=containers, host=host)
    if containers:
        app.invoke("run", containers=containers, host=host)
    else:
        app.invoke("up", host=host)


def run_formation(app, host, formation, task):
    """
    Common function to run a formation change.
    """
    run_hook = task.name != "Stopping containers"
    if run_hook:
        app.run_hooks(PluginHook.PRE_GROUP_START, host=host, formation=formation, task=task)
    container_in_error = None
    error_message = None
    show_tail_message = False
    try:
        FormationRunner(app, host, formation, task).run()
    # General docker/runner error
    except DockerRuntimeError as e:
        container_in_error = e.instance.container
        error_message = str(e)
        if e.code == "BOOT_FAIL":
            show_tail_message = True
    # An image was not found
    except ImageNotFoundException as e:
        container_in_error = e.instance.container
        error_message = "Missing image for {} - cannot continue boot.".format(container_in_error.name)

    # Set error properties to be used in post hook
    if run_hook:
        app.run_hooks(PluginHook.POST_GROUP_START, host=host, formation=formation, task=task,
                  container_in_error=container_in_error, error_message=error_message)
    if error_message:
        click.echo(RED(error_message))
        if show_tail_message:
            click.echo(CYAN("You can see its output with `bay tail {}`.".format(container_in_error.name)))
    else:
        task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)



import attr
import click
import sys

from .base import BasePlugin
from .gc import GarbageCollector
from ..cli.argument_types import HostType, ContainerType
from ..cli.colors import RED
from ..cli.table import Table
from ..cli.tasks import Task


@attr.s
class ImagesPlugin(BasePlugin):
    """
    Plugin for listing and destroying container images
    """

    requires = ["gc"]

    def load(self):
        self.add_command(image)


@click.group()
def image():
    """
    Allows operations on images.
    """
    pass


@image.command()
@click.option("--host", "-h", type=HostType(), default="default")
@click.pass_context
def list(ctx, host):
    """
    Lists available images _for containers_
    """
    if not ctx.invoked_subcommand:
        # Start a table
        table = Table([
            ("CONTAINER", 40),
            ("VERSION", 15),
            ("IMAGE", 50),
        ])
        table.print_header()
        # Iterate through docker containers and try to get their images
        for container in sorted(ctx.obj.containers, key=lambda c: c.name):
            versions = host.images.image_versions(container.image_name)
            for version, image_id in sorted(versions.items()):
                table.print_row([
                    container.name,
                    version,
                    image_id,
                ])


@image.command()
@click.option("--host", "-h", type=HostType(), default="default")
@click.option("--force", "-f", type=bool, default=False)
@click.argument("container", type=ContainerType())
@click.argument("version", default="latest")
@click.pass_obj
def destroy(app, host, container, version, force):
    """
    Removes an image for a container
    """
    # Check the version
    image_versions = host.images.image_versions(container.image_name)
    if version not in image_versions:
        click.echo(RED("There is no version '{} of {}".format(version, container.name)))
        sys.exit(1)
    # Run garbage collection on containers
    task = Task("Destroying {}:{}".format(container.image_name, version))
    garbage_collector = GarbageCollector(host)
    garbage_collector.gc_containers(task)
    # Destroy it
    host.client.remove_image(image_versions[version], force=force)
    task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)



import attr
import click

from .base import BasePlugin
from ..cli.argument_types import HostType
from ..cli.tasks import Task


@attr.s
class GcPlugin(BasePlugin):
    """
    Does garbage collection on demand.
    """

    provides = ["gc"]

    def load(self):
        self.add_command(gc)


@attr.s
class GarbageCollector:
    """
    Allows garbage collection on a host.
    """

    host = attr.ib()

    def gc_all(self, parent_task):
        task = Task("Running garbage collection", parent=parent_task)
        self.gc_containers(task)
        self.gc_networks(task)
        self.gc_images(task)
        task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

    def gc_containers(self, parent_task):
        """
        Remove all stopped containers
        """
        task = Task("Removing all stopped containers", parent=parent_task)
        response = self.host.client.prune_containers()
        task.finish(status="Done, reclaimed {}MB".format(
            response['SpaceReclaimed'] / 1024 / 1024), status_flavor=Task.FLAVOR_GOOD)

    def gc_networks(self, parent_task):
        """
        Remove all networks not used by at least one container
        """
        task = Task("Removing all networks not used by at least one container", parent=parent_task)
        self.host.client.prune_networks()
        task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)

    def gc_images(self, parent_task):
        """
        Remove all dangling images
        """
        task = Task("Removing all dangling images", parent=parent_task)
        response = self.host.client.prune_images({"dangling": True})
        task.finish(status="Done, reclaimed {}MB".format(
            response['SpaceReclaimed'] / 1024 / 1024), status_flavor=Task.FLAVOR_GOOD)


@click.command()
@click.option('--host', '-h', type=HostType(), default='default')
@click.pass_obj
def gc(app, host):
    """
    Runs the garbage collection manually.
    """
    GarbageCollector(host).gc_all(app.root_task)



import contextlib
import sys
import threading
import time


class ExceptionalThread(threading.Thread):
    """
    Thread subclass that allows exceptions to be easily re-raised in the parent.
    """

    def __init__(self, target=None, args=None, kwargs=None, daemon=False):
        threading.Thread.__init__(self, daemon=daemon)
        self.target = target
        self.args = args or tuple()
        self.kwargs = kwargs or {}
        self.__exception = None

    def run_with_exception(self):
        """
        This method should be overriden if you want to subclass this.
        """
        if self.target:
            self.target(*self.args, **self.kwargs)
        else:
            raise NotImplementedError("You must override run_with_exception")

    def run(self):
        """This method should NOT be overriden."""
        try:
            self.run_with_exception()
        except BaseException:
            self.__exception = sys.exc_info()

    def maybe_raise(self):
        if self.__exception is not None:
            raise self.__exception[1]


class ThreadSet(set):
    """
    Threadsafe set extended with check-and-set style operations
    """

    def __init__(self, *args, **kwargs):
        super(ThreadSet, self).__init__(*args, **kwargs)
        self.lock = threading.Lock()

    def check_and_add(self, value):
        """
        Checks if value is in the set and if it is not, adds it.
        Returns False if value is already in the set, True if value was not in the set and was added.
        """
        with self.lock:
            if value in self:
                return False
            self.add(value)
            return True

    @contextlib.contextmanager
    def entry_lock(self, value, interval=1):
        """
        Context manager that allows entry when the value is not in the set, and removes it
        once finished.
        """
        while not self.check_and_add(value):
            time.sleep(interval)
        yield
        self.remove(value)



import os

import attr
import click

from .base import BasePlugin
from .run import run_formation
from ..cli.argument_types import HostType
from ..cli.colors import CYAN, RED
from ..cli.table import Table
from ..cli.tasks import Task
from ..containers.profile import Profile, NullProfile
from ..docker.introspect import FormationIntrospector


@attr.s
class ProfilesPlugin(BasePlugin):
    """
    Plugin for managing and switching profiles.
    """

    provides = ["up"]

    def load(self):
        self.add_command(profile)
        self.add_command(up)
        self.add_command(list_profiles)


@click.command()
@click.argument('name', required=False)
@click.option("--host", "-h", type=HostType(), default="default")
@click.pass_obj
def profile(app, name, host):
    """
    Switch to a different profile, or list the active profile's name.
    """
    if name is None:
        # if no profile is provided, print current profile stack and exit
        if len(app.profiles) > 1:
            for i, parent_profile in enumerate(app.profiles[1:]):
                if i:
                    click.echo("%s↳ %s" % ("  " * i, parent_profile.name))
                else:
                    click.echo(parent_profile.name)
        else:
            click.echo(RED("No profile selected."))
        return

    # Find the profile they named on the command line
    parent_profile_path = os.path.join(
        app.config["bay"]["home"],
        "profiles",
        "{}.yaml".format(name)
    )
    if not os.path.isfile(parent_profile_path):
        click.echo(RED("Invalid profile name!"))
        return
    parent_profile = Profile(parent_profile_path)

    click.echo("Switching to profile %s" % CYAN(name))

    # Set it in the user profile, making a new user profile if needed
    if isinstance(app.user_profile, NullProfile):
        app.user_profile = Profile(app.user_profile_path, load_immediately=False)
    app.user_profile.parent_profile = name
    app.user_profile.save()


@click.command()
@click.option("--host", "-h", type=HostType(), default="default")
@click.pass_obj
def up(app, host):
    """
    Start up a profile by booting the default containers.
    Leaves any other containers that are running (shell, ssh-agent, etc.) alone.
    """
    profile = app.profiles[0].parent_profile if app.profiles else ''
    click.echo("Starting up profile %s..." % CYAN(profile))
    # Do removal loop first so we don't step on adding containers later
    formation = FormationIntrospector(host, app.containers).introspect()
    for instance in list(formation):
        # We remove all non-system containers, so that means ssh-agent and similar
        # containers will survive the process.
        if not instance.container.system and instance.formation:
            formation.remove_instance(instance)

    # Now add in containers
    for container in app.containers:
        if app.containers.options(container).get('default_boot'):
            formation.add_container(container, host)

    task = Task("Restarting containers", parent=app.root_task)
    run_formation(app, host, formation, task)


@click.command()
@click.option('--verbose/--quiet', '-v/-q', default=False)
@click.pass_obj
def list_profiles(app, verbose):
    """
    List all available profiles.
    """
    # the path where all profiles can be found
    profiles_home_dir = os.path.join(
        app.config['bay']['home'],
        "profiles",
    )
    corrupted_profiles = []
    if verbose:
        table = Table([
            ("PROFILE", 30),
            ("DESCRIPTION", 50),
        ])
        table.print_header()

    for filename in os.listdir(profiles_home_dir):
        if filename.endswith('.yaml'):
            profile_name = filename.split('.')[0]
            try:
                profile = Profile(os.path.join(profiles_home_dir, filename))
            except Exception as ex:
                # the profile is corrupted, the specifics of the error do not
                # matter in this case, where we are only trying to list
                # available profiles.
                corrupted_profiles.append(profile_name)
            else:
                if verbose:
                    table.print_row([
                        profile_name,
                        profile.description
                    ])
                else:
                    click.echo(profile_name)

    if corrupted_profiles:
        click.echo(
            "These profiles are defined but corrupted and cannot be loaded:"
        )
        click.echo(corrupted_profiles)



import attr


@attr.s
class NamedVolume:

    source = attr.ib()
    mode = attr.ib(default="rw")


@attr.s
class BoundVolume:

    source = attr.ib()
    mode = attr.ib(default="rw")
    required = attr.ib(convert=bool, default=True)


@attr.s
class DevMode:

    source = attr.ib()
    mode = attr.ib(default="rw")



import attr
import http.client
import ssl
import socket
import time
from docker.errors import NotFound

from .base import BasePlugin
from ..cli.tasks import Task
from ..constants import PluginHook
from ..exceptions import ContainerBootFailure, DockerRuntimeError


class WaitsPlugin(BasePlugin):
    """
    Contains the basic, standard waits. Waits' .check is called repeatedly and should return True if the condition is
    met or False if it is not.
    """

    provides = ["waits"]

    def load(self):
        self.add_hook(PluginHook.POST_RUN_CONTAINER, self.post_start)
        self.add_catalog_type("wait")
        self.add_catalog_item("wait", "http", HttpWait)
        self.add_catalog_item("wait", "https", HttpsWait)
        self.add_catalog_item("wait", "tcp", TcpWait)
        self.add_catalog_item("wait", "time", TimeWait)
        self.add_catalog_item("wait", "file", FileWait)

    def post_start(self, host, instance, task):
        # Loop through all waits and build instances
        wait_instances = []
        for wait in instance.container.waits:
            # Look up wait in app
            try:
                wait_class = self.app.get_catalog_items("wait")[wait["type"]]
            except KeyError:
                raise DockerRuntimeError(
                    "Unknown wait type {} for {}".format(wait["type"], instance.container.name)
                )
            # Initialise it and attach a task
            params = wait.get("params", {})
            params["instance"] = instance
            params["host"] = host
            wait_instance = wait_class(**params)
            wait_instance.task = Task("Waiting for {}".format(wait_instance.description()), parent=task)
            wait_instance.task.update(status="Waiting")
            wait_instances.append(wait_instance)

        # Check on them all until they finish
        while wait_instances:
            # See if the container actually died
            if not host.container_running(instance.name):
                task.update(status="Dead", status_flavor=Task.FLAVOR_BAD)

                raise ContainerBootFailure(
                    "Failed during waits",
                    instance=instance,
                )
            # Check the waits
            for wait_instance in list(wait_instances):
                try:
                    if wait_instance.ready():
                        wait_instance.task.finish(status="Done", status_flavor=Task.FLAVOR_GOOD)
                        wait_instances.remove(wait_instance)
                except Exception as e:
                    task.update(status="Failed", status_flavor=Task.FLAVOR_BAD)
                    raise DockerRuntimeError("Failed while waiting for {}:\n{}".format(instance.container.name, e))
            time.sleep(1)


@attr.s
class TcpWait:
    """
    Checks that a TCP port is open
    """

    instance = attr.ib()
    host = attr.ib()
    port = attr.ib(default=80)
    timeout = attr.ib(default=1)

    def ready(self):
        try:
            conn_kwargs = {}
            if self.timeout:
                conn_kwargs['timeout'] = self.timeout
            if self.port not in self.instance.port_mapping:
                raise DockerRuntimeError("Trying to wait on non-exposed port {}".format(self.port))
            conn = socket.create_connection(self.target(), **conn_kwargs)
            conn.close()
            return True
        except socket.error:
            return False

    def target(self):
        """
        Returns (host, port) target information.
        """
        if self.port not in self.instance.port_mapping:
            raise DockerRuntimeError("Trying to wait on non-exposed port {}".format(self.port))
        return (self.host.external_host_address, self.instance.port_mapping[self.port])

    def description(self):
        return "TCP on port {}".format(self.port)


@attr.s
class HttpWait(TcpWait):
    """
    Checks that a HTTP endpoint exists and returns a good value.
    """

    path = attr.ib(default="/")
    method = attr.ib(default="GET")
    headers = attr.ib(default=attr.Factory(dict))
    expected_codes = attr.ib(default=attr.Factory(lambda: range(200, 400)))

    connection_class = http.client.HTTPConnection

    def _get_connection(self, **kwargs):
        addr, port = self.target()
        return self.connection_class(addr, port, timeout=self.timeout, **kwargs)

    def _ready_request(self, conn):
        while True:
            try:
                conn.request(self.method, self.path, headers=self.headers)
                response = conn.getresponse()
                if response.status in self.expected_codes:
                    return True
                return False
            except http.client.RemoteDisconnected:
                conn.connect()
                continue

    def ready(self):
        conn = self._get_connection()
        # Run wait
        try:
            return self._ready_request(conn)
        except Exception:
            return False
        finally:
            conn.close()

    def description(self):
        return "HTTP on port {}".format(self.port)


@attr.s
class HttpsWait(HttpWait):
    """
    HTTPS variant of the HTTP wait.

    If verify_cert is True, this will let the underlying SSL library attempt to verify the CA
    cert provided by the server. This means the SSL library must be able to find the cert.
    Usually this is accomplished by setting the SSL_CERT_FILE environment variable.
    """
    verify_cert = attr.ib(default=False)

    connection_class = http.client.HTTPSConnection

    def _get_connection(self, **kwargs):
        context = ssl.create_default_context()
        # We are going to be making a request to an IP address, so we cannot rely on the cert
        # having the correct hostname.
        context.check_hostname = False
        if not self.verify_cert:
            context.verify_mode = ssl.CERT_NONE
        return super(HttpsWait, self)._get_connection(context=context)

    def ready(self):
        conn = self._get_connection()
        try:
            return self._ready_request(conn)
        except (ssl.SSLError, ssl.CertificateError):
            # If there is a problem with the cert or SSL connection, error out immediately
            self.task.update(status="SSL error")
            raise
        except Exception:
            return False
        finally:
            conn.close()

    def description(self):
        return "HTTPS on port {}".format(self.port)


@attr.s
class TimeWait:
    """
    Waits a number of seconds
    """

    # Everything needs instance and host, alas
    instance = attr.ib()
    host = attr.ib()
    seconds = attr.ib()

    def __attrs_post_init__(self):
        self.wait_until = time.time() + int(self.seconds)

    def ready(self):
        return time.time() >= self.wait_until

    def description(self):
        return "{} seconds".format(self.seconds)


@attr.s
class FileWait:
    """
    Waits until a named file appears in the container
    """

    instance = attr.ib()
    host = attr.ib()
    path = attr.ib()
    waiting_name = attr.ib(default=None)

    def ready(self):
        # Get the file from the host
        try:
            self.host.client.get_archive(self.instance.name, self.path)
        except NotFound:
            return False
        else:
            return True

    def description(self):
        return self.waiting_name or "file {}".format(self.path)



def file_size(value, fmt="{value:.1f} {suffix}", si=False):
    """
    Takes a raw number of bytes and returns a humanized filesize.
    """
    if si:
        base = 1000
        suffixes = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    else:
        base = 1024
        suffixes = ("B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")
    max_suffix_index = len(suffixes) - 1
    for i, suffix in enumerate(suffixes):
        unit = base ** (i + 1)
        if value < unit or i == max_suffix_index:
            return fmt.format(value=(base * value / unit), suffix=suffix)



import attr
import click

from .base import BasePlugin
from ..cli.argument_types import HostType
from ..cli.table import Table
from ..docker.introspect import FormationIntrospector
from ..utils import humanize


@attr.s
class PsPlugin(BasePlugin):
    """
    Plugin to see what's running right now.
    """

    def load(self):
        self.add_command(ps)


@click.command()
@click.option("--host", "-h", type=HostType(), default="default")
@click.option("--stats/--no-stats", "-s")
@click.pass_obj
def ps(app, host, stats):
    """
    Shows details about all containers currently running
    """
    # Run the introspector to get the details
    formation = FormationIntrospector(host, app.containers).introspect()
    # Print formation details
    if stats:
        table = Table([
            ("NAME", 30),
            ("DOCKER NAME", 40),
            ("MEMORY", 10),
            ("PORTS (CONTAINER->HOST)", 30),
        ])
    else:
        table = Table([
            ("NAME", 30),
            ("DOCKER NAME", 40),
            ("PORTS (CONTAINER->HOST)", 30),
        ])
    table.print_header()
    for instance in sorted(formation, key=lambda i: i.name):
        row = [
            instance.container.name,
            instance.name,
        ]
        if stats:
            # Get the memory usage from the docker host
            stats = host.client.stats(instance.name, decode=True, stream=False)
            row.append(humanize.file_size(stats['memory_stats']['usage']))
        # Add in port info
        row.append(", ".join(
            "{}->{}".format(private, public)
            for private, public in instance.port_mapping.items()
        ))
        table.print_row(row)



import attr
from ..containers.formation import ContainerFormation, ContainerInstance
from ..exceptions import DockerRuntimeError

import warnings


@attr.s
class FormationIntrospector:
    """
    Given a docker host, introspects it to work out what Formation it is
    currently running and returns that for use/comparison with a desired new
    Formation.
    """
    host = attr.ib()
    graph = attr.ib()
    network = attr.ib(default=None)
    formation = attr.ib(init=False)

    class ContainerNotFound(DockerRuntimeError):
        pass

    def __attrs_post_init__(self):
        if self.network is None:
            self.network = self.graph.prefix

    def introspect(self):
        """
        Runs the introspection and returns a ContainerFormation.
        """
        # Make the formation
        self.formation = ContainerFormation(self.graph, self.network)
        # Go through all containers on the remote host that are running and on the right network
        for container in self.host.client.containers(all=False):
            if self.network in container['NetworkSettings']['Networks']:
                self.add_container(container['Names'][0].lstrip("/"))
        # As a second phase, go through and resolve links
        for instance in self.formation:
            instance.resolve_links()
        return self.formation

    def introspect_single_container(self, name):
        """
        Returns a single container introspected directly.
        """
        # Inspect image and list images have different formats, so we use list with filter here to match the other code
        details = self.host.client.containers(filters={"name": name})
        if not details:
            raise DockerRuntimeError("Cannot introspect single container {}".format(name))

        # A race condition in docker [2017/09] means it returns a different format for containers that have just died
        if isinstance(details[0], dict):
            container_name = details[0]["Names"][0].lstrip("/")
        else:
            container_name = details[0]

        return self._create_container(container_name)

    def add_container(self, container_details):
        try:
            instance = self._create_container(container_details)
            self.formation.add_instance(instance)
        except self.ContainerNotFound as e:
            warnings.warn(e.args[0])

    def _create_container(self, container_name):
        """
        Returns a container build from introspected information
        """
        assert isinstance(container_name, str)
        container_details = self.host.client.inspect_container(container_name)
        # Find the container name in the graph
        try:
            labels = container_details['Config']['Labels']
            # Use the bay-specific (not eventbrite-specific, just named uniquely as per the docker label spec) label
            # to work out what container name this was.
            container = self.graph[labels['com.eventbrite.bay.container']]
        except KeyError:
            raise self.ContainerNotFound(
                (
                    "Cannot find local container for running container {}. "
                    "Perhaps its configuration was moved or deleted?"
                ).format(container_name)
            )
        # Get the image hash
        image = container_details['Image']
        assert ":" in image
        if image.startswith("sha256:"):
            image_id = image
        else:
            # It's a string name of a image
            # CONVERT IMAGE NAME INTO HASH USING REPO
            name, tag = image.split(":", 1)
            image_id = self.host.images.image_version(name, tag)
        # Work out links
        links = {}
        for link in (container_details['NetworkSettings']['Networks'][self.network].get('Links', None) or []):
            linked_container_name, link_alias = link.split(":", 1)
            links[link_alias] = linked_container_name
        # Work out devmodes
        mounted_targets = set()
        devmodes = set()
        for mount in container_details['Mounts']:
            mounted_targets.add(mount['Destination'])
        for devmode, mounts in container.devmodes.items():
            if all((destination in mounted_targets) for destination in mounts.keys()):
                devmodes.add(devmode)
        # Make a formation instance
        instance = ContainerInstance(
            name=container_name,
            container=container,
            image_id=image_id,
            links=links,
            devmodes=devmodes,
        )
        # Set extra networking attributes because it's running
        instance.ip_address = container_details['NetworkSettings']['Networks'][self.network]['IPAddress']
        instance.port_mapping = {}
        for container_port, host_details in container_details['NetworkSettings'].get('Ports', {}).items():
            if host_details:
                private_port = int(container_port.split("/", 1)[0])
                public_port = int(host_details[0]['HostPort'])
                instance.port_mapping[private_port] = public_port
        return instance



import sys
import click

from .base import BasePlugin
from ..cli.argument_types import HostType, ContainerType
from ..cli.colors import RED


class TailPlugin(BasePlugin):
    """
    Plugin to let you view container output.
    """

    provides = ["tail"]

    def load(self):
        self.add_command(tail)


@click.command()
@click.option("--host", "-h", type=HostType(), default="default")
@click.option('--follow/--no-follow', '-f', default=False)
@click.argument("container", type=ContainerType())
@click.argument("lines", default="10")
@click.pass_obj
def tail(app, host, container, lines, follow=False):
    """
    Shows logs from a container. Optional second argument specifies a number of lines to print, or "all".
    """
    # We don't use formation here as it doesn't include stopped containers;
    # instead, we manually go through the list.
    for docker_container in host.client.containers(all=True):
        if docker_container['Labels'].get('com.eventbrite.bay.container', None) == container.name:
            # We found it!
            container_name = docker_container['Names'][0]
            break
    else:
        click.echo(RED("Cannot find instance of {} to print logs for.".format(container.name)))
        sys.exit(1)
    # Either stream or just print directly
    if lines != "all":
        try:
            lines = int(lines)
        except Exception as ex:
            click.echo(RED("Invalid number of lines: {}".format(lines)))
            sys.exit(1)
    if follow:
        for line in host.client.logs(container_name, tail=lines, stream=True):
            click.echo(line, nl=False)
    else:
        click.echo(host.client.logs(container_name, tail=lines))
